package net.javase.db.utils;

import net.javase.db.annotation.Column;
import net.javase.db.annotation.Entity;
import net.javase.db.annotation.Id;
import net.javase.db.config.DbConfig;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.sql.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.beanutils.PropertyUtils;

/**
 * DbUtil
 *
 * @author Frank
 */
public class DbUtil {

    private static Connection conn;

    static {
        try {
            Class.forName(DbConfig.get().getDriver());
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }
    }

    /**
     * 获取数据库连接对象
     *
     * @return {@link Connection}
     */
    public static Connection getConnection() {
        // 获取连接并捕获异常
        try {
            if (conn == null || conn.isClosed())
                conn = DriverManager.getConnection(
                        DbConfig.get().getUrl(),
                        DbConfig.get().getUsername(),
                        DbConfig.get().getPassword());
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return conn;// 返回连接对象
    }

    /**
     * 关闭数据库连接
     *
     * @param conn {@link Connection} 数据库连接
     */
    public static void closeConn(Connection conn) {
        try {
            if (conn != null && !conn.isClosed())
                conn.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * 关闭数据库连接。
     *
     * @param conn 数据库连接
     * @param stmt Statement对象
     * @param rs   结果集
     */
    public static void closeAll(Connection conn, Statement stmt, ResultSet rs) {
        // 若结果集对象不为空，则关闭
        try {
            if (rs != null && !rs.isClosed())
                rs.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
        // 若Statement对象不为空，则关闭
        try {
            if (stmt != null && !stmt.isClosed())
                stmt.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
        // 若数据库连接对象不为空，则关闭
        try {
            if (conn != null && !conn.isClosed())
                conn.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * 增、删、改操作
     *
     * @param sql    sql语句
     * @param params 参数数组
     * @return 执行结果
     */
    public static int executeUpdate(String sql, Object... params) {
        int result = 0;
        conn = getConnection();
        PreparedStatement pstmt = null;
        try {
            pstmt = conn.prepareStatement(sql);
            for (int i = 0; i < params.length; i++) {
                pstmt.setObject(i + 1, params[i]);
            }
            result = pstmt.executeUpdate();
        } catch (SQLException e) {
            e.printStackTrace();
        } finally {
            closeAll(null, pstmt, null);
        }
        return result;
    }

    /**
     * 查询操作
     *
     * @param sql    sql语句
     * @param params 参数数组
     * @return 查询结果集
     */
    public static ResultSet executeQuery(String sql, Object... params) {
        conn = getConnection();
        PreparedStatement pstmt = null;
        ResultSet rs = null;
        try {
            pstmt = conn.prepareStatement(sql);
            for (int i = 0; i < params.length; i++) {
                pstmt.setObject(i + 1, params[i]);
            }
            rs = pstmt.executeQuery();
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return rs;
    }

    /**
     * 查询实体列表
     * 
     * @param <T>    T
     * @param clazz  entity class
     * @param sql    查询sql
     * @param params sql参数
     * @return entity list
     */
    public static <T> List<T> selectList(Class<T> clazz, String sql, Object... params) {
        conn = getConnection();
        PreparedStatement pstmt = null;
        ResultSet rs = null;
        List<T> list = new ArrayList<T>();
        try {
            pstmt = conn.prepareStatement(sql);
            for (int i = 0; i < params.length; i++) {
                pstmt.setObject(i + 1, params[i]);
            }
            rs = pstmt.executeQuery();
            Map<String, String> entityMateData = getMetaData(clazz);
            ResultSetMetaData rsmd = rs.getMetaData();
            int columnCount = rsmd.getColumnCount();
            while (rs.next()) {
                T obj = clazz.newInstance();
                for (int i = 1; i <= columnCount; i++) {
                    String columnName = rsmd.getColumnName(i);
                    Object value = rs.getObject(i);
                    try {
                        if (entityMateData.get(columnName) != null) {
                            PropertyUtils.setProperty(obj, entityMateData.get(columnName), value);
                        }
                    } catch (InvocationTargetException e) {
                        e.printStackTrace();
                    } catch (NoSuchMethodException e) {
                        e.printStackTrace();
                    }
                }
                list.add(obj);
            }
            entityMateData = null;
        } catch (SQLException | InstantiationException | IllegalAccessException e) {
            throw new RuntimeException(e);
        } finally {
            closeAll(conn, pstmt, rs);
        }
        return list;
    }

    /**
     * 获取entity对应的column对应关系
     * 
     * @param <T>   T
     * @param clazz entity class
     * @return map
     */
    private static <T> Map<String, String> getMetaData(Class<T> clazz) {
        Map<String, String> map = new HashMap<>();
        Field[] fields = clazz.getDeclaredFields();
        for (Field field : fields) {
            Column c = field.getAnnotation(Column.class);
            if (c != null) {
                if (StrUtil.notBlank(c.name())) {
                    map.put(c.name(), field.getName());
                }
            }
            map.put(field.getName(), field.getName());
        }
        return map;
    }

    /**
     * 生成建表语句
     *
     * @param clazz     实体类
     * @param underline 是否下划线命名
     * @param <T>       实体类
     * @return sql
     */
    public static <T> String createSchemaSql(Class<T> clazz, boolean underline) {
        String table = clazz.getSimpleName();
        Entity e = clazz.getAnnotation(Entity.class);
        if (e != null && !StrUtil.isBlank(e.table())) {
            table = e.table();
        } else {
            table = StrUtil.camelToUnderscore(table);
        }
        StringBuilder sqlBuilder = new StringBuilder();
        sqlBuilder.append("CREATE TABLE IF NOT EXISTS `").append(table).append("` (");
        Field[] fields = clazz.getDeclaredFields();
        for (Field field : fields) {
            String columnName = field.getName();
            String jdbcType = ""; // 类型
            boolean notNull = false; // 是否为空
            String defaultValue = ""; // 默认值
            boolean isKey = false;

            Id id = field.getAnnotation(Id.class);
            if (id != null) {
                isKey = true;
                notNull = true;
            }

            Column c = field.getAnnotation(Column.class);
            if (c != null) {
                if (StrUtil.notBlank(c.name())) {
                    columnName = c.name();
                }
                notNull = c.notNull();
                defaultValue = c.defaultValue();
            }

            Class<?> fieldType = field.getType();
            if (String.class.equals(fieldType)) {
                jdbcType = "VARCHAR(255)";
            } else if (Integer.class.equals(fieldType) || int.class.equals(fieldType)) {
                jdbcType = "INT";
            } else if (Long.class.equals(fieldType) || long.class.equals(fieldType)) {
                jdbcType = "BIGINT";
            } else if (Boolean.class.equals(fieldType) || boolean.class.equals(fieldType)) {
                jdbcType = "BOOLEAN";
            } else if (Double.class.equals(fieldType) || double.class.equals(fieldType)) {
                jdbcType = "DOUBLE";
            } else if (Date.class.equals(fieldType) || java.util.Date.class.equals(fieldType)) {
                jdbcType = "datetime";
            } else if (Float.class.equals(fieldType) || float.class.equals(fieldType)) {
                jdbcType = "float";
            } else {
                jdbcType = "VARCHAR(255)";
            }

            if (underline) {
                columnName = StrUtil.camelToUnderscore(columnName);
            }

            sqlBuilder.append("`").append(columnName).append("` ")
                    .append(jdbcType).append(" ")
                    .append(notNull ? "NOT NULL" : "NULL").append(" ")
                    .append(StrUtil.notBlank(defaultValue) ? "DEFAULT '" + defaultValue + "'" : "").append(" ")
                    .append(isKey ? "PRIMARY KEY AUTO_INCREMENT" : "")
                    .append(",");
        }
        // 去掉最后一个逗号
        if (fields.length > 0) {
            sqlBuilder.deleteCharAt(sqlBuilder.length() - 1);
        }

        sqlBuilder.append(") ENGINE=InnoDB DEFAULT CHARSET=utf8;");
        return sqlBuilder.toString();
    }
}
